/***********************************************************************
|  Enhanced C library
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

#ifndef ECLIB_HEADER
/* don't include this file more than once */
#define ECLIB_HEADER 1

#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>

#ifndef O_BINARY
#define O_BINARY 0
#endif
#ifndef O_TEXT
#define O_TEXT 0
#endif
#ifndef O_SYNC
#define O_SYNC 0
#endif

/* values for structure allocation */
#define TUPLEN 240 /* longest tuple, largest structure */
#define AVAIL_CONSTANT 0x5a3c74ae /* word indicating structure is available */
#define NULLKEY 0L
#define MAX_ALEV 4
#define BYTES_MALLOC 8180
#define K_MALLOC 8
#define init_0(x) memset(x, 0, sizeof(*(x)))

#define SCAN_NL_STRING 1
#define SCAN_LONG_STRING 2
#define SCAN_FSEEK 3
#define SCAN_FILE_ACCESS 4
#define SCAN_STREAM_LIMIT 5
#define SCAN_EOF_CONTEXT 6
#define SCAN_EOF_CMT 7
#define SCAN_EOF_STRING 8
#define SCAN_NESTED_CMT 9
#define SCAN_BAD_DIGIT 10
#define SCAN_PRECISION_LONG 11
#define SCAN_BADCHAR 12
#define SCAN_LONGNAME 13

#define NODE_BAD_LEVEL 20
#define NODE_ALLOC_SIZE 21
#define NODE_NOMEM_TOTAL 22
#define NODE_NOMEM_STRUCT 23
#define NODE_ALLOC_FAILURE 24
#define NODE_LEVEL_STRUCT_MISMATCH 25
#define NODE_CORRUPT 26
#define NODE_NULL_PTR 27
#define NODE_NONEMPTY 28
#define NODE_DUPLICATE 29
#define NODE_LIST_PTR 30
#define NODE_SERIAL_DIRS_0 31
#define NODE_SERIAL_DIRS_2 32
#define NODE_SERIAL_LIST_0 33
#define NODE_MKLOCK 35
#define NODE_LOCKED 36
#define NODE_SERIAL_RANGE 37
#define NODE_STRUCT_TYPE 38
#define NODE_SERIAL_64K 39
#define NODE_TOO_LARGE 40
#define NODE_RDONLY 41

typedef unsigned short str_code;

/* some externs */
extern char *code_to_string();
extern str_code string_to_code();
extern FILE *efopen();
extern long elseek(), filesize();
extern int errno;
extern char *emalloc();

extern struct STRUCT_DESC {
char *name;
unsigned short size;
short initchar;
unsigned char minlev, maxlev;
long memlimit, base;
char norders, nodetype, fd;
/* boolean attributes */
char holdopen, rdonly, noholes;
} struct_desc[];
extern short ZC_max_struct;

extern struct IDX_DESC {
char relnum;
char cls; /* char, short, long */
char fd;
char holdopen;
char *member;
short offset;
short numer, denom; /* numerator-denominator, packing ratio */
char bounded;
char filler;
} idx_desc[];

/* the serial list control structure for in-memory lists */
struct serial_inmem {
char *inmem; /* point to the data; memory only */
char **inmem_ptr; /* update when we change inmem after a realloc() */
long base;
short n_nodes, n_room;
char relnum; /* relation number */
char setup; /* this serial list set up */
short filler;
};

/* cursor, to loop through a relation in sequence or by index */
/* set copy, serial_inmem, relnum, and possibly the key */
struct rel_cursor {
struct serial_inmem *x; /* when the relation is in memory */
char *copy; /* copy record instead of just returning a pointer */
long lowkey, highkey, rowid;
char relnum, idxnum, eof, filler;
short fd, ifd;
};

/* variable length integer */
struct VINT {
short size;
short waste;
unsigned short value[1];
};

/* structure of a polynomial */
#define MAXVAR_TERM 8 /* max number of variables in the term of a polynomial */
struct POLY {
struct POLY *next; /* chain top and bottom of a rational function */
short nterms;
struct TERM {
struct VINT *coef; /* coefficient on term */
char nvars; /* number of variables in term */
char degree; /* degree of term */
char vars[MAXVAR_TERM]; /* variables in term */
char exps[MAXVAR_TERM]; /* exponents on variables */
} terms[1]; /* variable length array */
};

/* fixed integers */
extern struct VINT izero, ione, ibase, obase, digit, *modulus;

/* fixed polynomials */
extern struct POLY pzero, pone, pconst, pvar;
/* each letter is potentially a variable */
extern struct POLY *vb[26];
extern char vbr[26];
extern struct POLY *dotval; /* what was last displayed */

/* error messages */
extern char er_const[], er_arg[], er_resrc[], er_exp[],
	er_zero[], er_lhs[];

/* global flags */
extern char intdiv, is_square, err_seen;

/* external functions for infinite precision arithmetic operations */
extern struct VINT *newnum(), *string_to_vint(), *long_to_vint();
extern struct POLY *newpoly();
extern bcfree(), allfree(), polyfree();
extern struct VINT *iadd(), *isub(), *imul(), *idiv(), *irem();
extern struct VINT *igcd(), *isqr();
extern void ishr(), ishl();
extern struct POLY *padd(), *pmul();
extern struct POLY *radd(), *rsub(), *rmul(), *rdiv(), *rrem();
extern struct VINT *vexp(); /* exponentiation */
extern struct VINT *vbang(); /* factorial */
extern struct VINT *vchoose(); /* binary coefficients */
extern struct POLY *rexp(), *rchoose(); /* as above, but taking polynomials */
extern struct POLY *polysimp(); /* simplify polynomials */
extern struct VINT *modinv(); /* modular inverse */

/* Dates are packed into a long.  Year*367 + day_within_year.
 * Days are numbered from 1 to 365.
 * A day of 0 means an indeterminant date,
 * with only the year specified.
 * When sorted, indeterminant dates appear first (0),
 * followed by the actual dates.
 * You cannot accurately subtract two dates to get number of days between,
 * because of the 0-day dates, and the 2/29 holes,
 * which are present in every year, but only valid every four years.
 * Use the routine diff_dates() instead. */
typedef long date;
#define YEAR_DAY_RATIO 367L
#define EARLY_DATE (-9999L*YEAR_DAY_RATIO)
extern char *datestring();
extern date datecv(), add_days(), string_to_date(), current_date();
extern int dirstart; /* begin looping through the files in a directory */
extern char *nextfile();

extern int (*ZC_fn_uncache)();
extern char *ZC_get_struct(), *ZC_copy_struct();
extern char *ZC_select_row_mem(), *ZC_select_row_file();
extern long ZC_insupd_row_mem(), ZC_insupd_row_file();
extern long max_serial();
extern char *ZC_ht_insert(), *ZC_ht_rm();

extern char *strchr(), *strrchr(), *getenv();

#endif
