/***********************************************************************
|  Interactively configurable forms
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

/* eval.c: evaluate a formula */

#include "icfgram.h"

static char *typename[] = {
"???","money","number","string","string","boolean","date","domain member"};
static long *lalist; /* loop array list */
static int lacount; /* size of lalist */
static int laindex; /* loop array index */
static int lamax; /* upper bound for laindex */
static short laend; /* block number that ends loop array */
static long stack[2*STACKSIZE];
static int stackptr;
static short cbn; /* current block number */

static badcv(type1, type2)
{
eprint("Cannot convert from type %d to type %d", type1, type2);
} /* badcv */

/* convert value to symbol in a domain, or vice versa */
cv_dom_mm(domnum, elnum, elsym, ia)
short domnum, elnum;
str_code elsym;
{
struct domain_member m, *m1;
struct domain d, *d1;

d.dnum = domnum;
d1 = find_domain(&domains, &d);
if(!elnum) {
m.elsym = elsym;
m1 = find_domain_member_1(&d1->members, &m);
if(m1) return m1->elnum;
}
if(elnum) {
m.elnum = elnum;
m1 = find_domain_member(&d1->members, &m);
if(m1) return m1->elsym;
}
if(ia) printf("\n%s\n", code_to_string(d1->helpmsg));
return 0;
} /* cv_dom_mm */

/* convert date to string */
static long cv_date_string(l)
date l;
{
l = string_to_code(datestring(l));
return l;
} /* cv_date_string */

/* convert string to date */
static long cv_string_date(l)
long l;
{
str_code sc = l;
char *s = code_to_string(sc);
l = string_to_date(s);
if(!l) eprint("Cannot convert string %s to a date", s);
return l;
} /* cv_string_date */

/* convert string to int */
static long cv_string_int(l)
long l;
{
str_code sc = l;
char *s = code_to_string(sc);
if(zc_strtol(s, &s, &l) || *s)
eprint("Cannot convert string to integer");
return l;
} /* cv_string_int */

/* convert from int to bool */
static long cv_int_bool(l)
long l;
{
if(l) l = 1;
return l;
} /* cv_int_bool */

/* convert from string to bool */
static long cv_string_bool(l)
long l;
{
str_code sc = l;
char *s = code_to_string(sc);
if(s[0] && !s[1]) { /* ok so far */
if(s[0] == 'y' || s[0] == 'Y')
return 1L;
if(s[0] == 'n' || s[0] == 'N')
return 0L;
}
return cv_int_bool(cv_string_int(l));
} /* cv_string_bool */

/* convert int to string */
static long cv_int_string(l)
long l;
{
char s[11];
sprintf(s, "%ld", l);
l = string_to_code(s);
return l;
} /* cv_int_string */

/* convert bool to string */
static long cv_bool_string(l)
long l;
{
l = string_to_code(l ? "Y" : "N");
return l;
} /* cv_string_bool */

/* force an item to become boolean */
static cv_bool(l, type)
long l;
short type;
{
switch(type) {
case TYPE_MONEY: case TYPE_INT:
l = cv_int_bool(l);
break;
case TYPE_STRING:
l = cv_string_bool(l);
case TYPE_BOOL:
break;
case TYPE_DATE:
badcv(type, 5);
default:
eprint("@default1[%d] in eval.c", type);
} /* end switch */
return (int)l;
} /* cv_bool */

static str_code cv_string(l, type)
long l;
short type;
{
switch(type) {
case TYPE_MONEY:
case TYPE_INT:
l = cv_int_string(l);
case TYPE_STRING:
break;
case TYPE_BOOL:
l = cv_bool_string(l);
break;
case TYPE_DATE:
l = cv_date_string(l);
break;
default:
eprint("@default6[%d] in eval.c", type);
} /* end switch */
return (str_code)l;
} /* cv_string */

/* force an item to become a domain member */
static long cv_domain(l, type, domnum)
long l;
short type, domnum;
{
short domval; /* new domain value */

domval = l;
if(type == domnum) return domval;

if(type >= 100) { /* different domain */
l = cv_dom_mm(type, domval, 0, 0);
type = 4;
}

switch(type) {
case TYPE_INT:
l = cv_int_string(l);
case TYPE_STRING:
break;
case TYPE_DATE:
case TYPE_MONEY:
case TYPE_BOOL:
badcv(type, 100);
default:
eprint("@default2[%d] in eval.c", type);
} /* end switch */
domval = cv_dom_mm(domnum, 0, (str_code)l, 0);
if(!domval)
eprint("%s is not a member of the domain %d",
code_to_string((str_code)l), domnum);
return domval;
} /* cv_domain */

/* evaluate binary operation */
static evalbin(lp1, op, lp2, lp3)
long *lp1, *lp2, *lp3;
short op;
{
short type1 = lp1[0];
short type2 = lp2[0];
long arg1 = lp1[1];
long arg2 = lp2[1];

if(icfdebug) {
/* don't print first arg if unary operator */
if(op < TOK_MAX && op > TOK_AVG && op != TOK_NOT && op != TOK_ASG)
printval(stdout, arg1, type1);
printf(" %s ",ZC_tok_strings[op]);
printval(stdout, arg2, type2);
printf(" -> ");
}

if(op == TOK_USING) {
usingconv(arg1, type1, (str_code)arg2, lp3);
return;
}

if(op >= TOK_MAX && op <= TOK_AVG) {
aggrigate((str_code)arg2, op, lp3);
return;
}

lp3[0] = 5;

/* first deal with domains */
if(type1 >= 100) {
/* left is a domain, convert it to string,
 *  unless we are making an assignment or comparison */
if(op == TOK_ASG || op == TOK_EQ || op == TOK_NE) {
arg2 = cv_domain(arg2, type2, type1);
goto mustintcompare;
}
if(arg1) {
/* could be 0 when op = TOK_NOT (unary) */
arg1 = cv_dom_mm(type1, (short)arg1, 0, 0);
type1 = 4;
}
} /* first is domain */

if(type2 >= 100) {
arg2 = cv_dom_mm(type2, (short)arg2, 0, 0);
type2 = 4;
}

if(op == TOK_OR || op == TOK_AND) {
arg1 = cv_bool(arg1, type1);
arg2 = cv_bool(arg2, type2);
if(op == TOK_OR)
lp3[1] = arg1 | arg2;
else
lp3[1] = arg1 & arg2;
return;
}

if(op == TOK_NOT) {
lp3[1] = ! cv_bool(arg2, type2);
return;
}

if(op == TOK_PIPE) { /* concatenate */
concat(cv_string(arg1, type1), cv_string(arg2, type2), lp3);
return;
}

/* if the first is a date, the second (usually) must become a date */
if(type1 == TYPE_DATE) {
if(type2 == 1 || type2 == 5) goto badcvop;

if(type2 == TYPE_DATE) { /* both are dates */
if(op == TOK_SUB) { /* subtract two dates */
lp3[0] = 2;
lp3[1] = diff_dates(arg1, arg2);
return;
}

mustintcompare:
if(op == TOK_ASG) {
make_asg:
lp3[0] = type1;
lp3[1] = arg2;
return;
}

switch(op) {
case TOK_EQ:
lp3[1] = (arg1 == arg2);
break;
case TOK_NE:
lp3[1] = (arg1 != arg2);
break;
case TOK_LT:
lp3[1] = (arg1 < arg2);
break;
case TOK_LE:
lp3[1] = (arg1 <= arg2);
break;
case TOK_GT:
lp3[1] = (arg1 > arg2);
break;
case TOK_GE:
lp3[1] = (arg1 >= arg2);
break;
default:
goto badcvop;
} /* end switch */
return;
} /* both are dates */

if(op == TOK_ADD || op == TOK_SUB) {
if(type2 == 4)
arg2 = cv_string_int(arg2);
lp3[0] = TYPE_DATE;
lp3[1] = add_days(arg1, (int)arg2);
return;
} /* date +- */

if(type2 != 4) goto badcvop;
arg2 = cv_string_date(arg2);

goto mustintcompare;
} /* first is date */

if(type2 == TYPE_DATE) { /* second is date, first is not */
if(type1 == 1 || type1 == 5) goto badcvop;

if(op == TOK_ADD) {
if(type1 == 4)
arg1 = cv_string_int(arg1);
lp3[0] = TYPE_DATE;
lp3[1] = add_days(arg2, (int)arg1);
return;
} /*  days + date */

/* only relops left */
if(type1 != 4) goto badcvop;
arg1 = cv_string_date(arg1);
goto mustintcompare;
} /* second is date */

if(type1 == 5) { /* first is bool */
if(op >= TOK_ADD && op <= TOK_MUL && (type2 <= 2 || type2 == 5))
type1 = 2;
else {
arg2 = cv_bool(arg2, type2);
goto mustintcompare;
}
} /* first is bool */

if(type2 == 5) {
if(op >= TOK_ADD && op <= TOK_MUL && type1 <= 2)
type2 = 2;
else {
if(op == TOK_ASG) {
switch(type1) {
case 4:
arg2 = cv_bool_string(arg2);
case 5:
case 2:
case 1:
break;
default:
goto badcvop;
} /* switch */
goto make_asg;
} /* assign */
arg1 = cv_bool(arg1, type1);
goto mustintcompare;
}
} /* second is bool */

/* now both are strings or numbers (possibly money numbers) */
/* if one is string and the other is number,
 * we switch to strings for relational operators,
 * and to numbers for arithmetic operators */
if(op >= TOK_EQ && op <= TOK_GE) { /* relational */
if(type1 == 4 && type2 <= 2) {
arg2 = cv_int_string(arg2);
type2 = 4;
}
if(type2 == 4 && type1 <= 2) {
arg1 = cv_int_string(arg1);
type1 = 4;
}
if(type1 == 4) {
char *s1 = code_to_string((str_code)arg1);
char *s2 = code_to_string((str_code)arg2);
arg1 = strcmp(s1, s2);
arg2 = 0;
}
goto mustintcompare;
} /* relational */

if(op == TOK_ASG) { /* assignment */
if(type1 == 4 && type2 <= 2) {
arg2 = cv_int_string(arg2);
type2 = 4;
}
if(type1 <= 2 && type2 == 4) {
arg2 = cv_string_int(arg2);
}
goto make_asg;
} /* assign */

/* only arithmetics left */
if(type1 == 4) arg1 = cv_string_int(arg1);
if(type2 == 4) arg2 = cv_string_int(arg2);
lp3[0] = 2;
if(type1 == 1 || type2 == 1) lp3[0] = 1;
switch(op) {
case TOK_ADD:
lp3[1] = arg1 + arg2;
break;
case TOK_SUB:
lp3[1] = arg1 - arg2;
break;
case TOK_MUL:
if(arg1 && 0x7fffffffL/arg1 <= arg2)
eprint("integer overflow");
lp3[1] = arg1 * arg2;
break;
case TOK_DIV:
if(!arg2) eprint("division by 0");
lp3[1] = arg1 / arg2;
break;
case TOK_REM:
if(!arg2) eprint("division by 0");
lp3[1] = arg1 % arg2;
break;
default:
goto badcvop;
} /* switch */

return;

badcvop:
if(type1 >= 100) type1 = 7;
if(type2 >= 100) type2 = 7;
eprint("Cannot perform operation %s on a %s and a %s",
ZC_tok_strings[op], typename[type1], typename[type2]);
} /* evalbin */

/* evaluate a formula */
eval_form(b)
struct block *b;
{
long *exp = b->formula;
short esize = b->fmsize;
int i;
long fm[EXPLEN*2];
long *lp;

if(b->fromstack == 2) {
if(stack[stackptr-2] == 99)
eprint("Not enough parameters passed or values returned");
fm[1] = stack[--stackptr];
fm[0] = stack[--stackptr];
} else {

/* make a copy */
memcpy(fm, exp, 8*esize);

/* replace variables with their values */
lp = fm;
for(i=0; i<esize; ++i, lp += 2) {
if(lp[0] == EXP_VARNUM) {
short vt;
var_value((str_code)lp[1], lp+1, &vt);
lp[0] = vt;
}
} /* substitute variables */

while(esize > 1) {
lp = fm;
for(i=0; i<esize-2; ++i, lp += 2) {
if(lp[4] != EXP_OP) continue;
if(lp[0] == EXP_OP) continue;
if(lp[2] == EXP_OP) continue;
break; /* found the sequence argument argument operator */
}
if(i == esize-2)
eprint("@bad prefix format in formula");

evalbin(lp, (short)lp[5], lp+2, lp+4);
if(icfdebug) {
printval(stdout, lp[5], (short)lp[4]);
nl();
}
memcpy(lp, lp+4, (esize-i-2)*8);
#ifdef COHERENT
/* lines needed to get around coherent memcpy/short/cast bug */
{
short coh_bug;
coh_bug = (short)(lp[0]);
coh_bug = (short)(lp[1]);
}
#endif
esize -= 2;
} /* end collapsing polish formula */
} /* pop or formula */

/* if we have an assignment then there is one more binary op to run */
if(b->blktype == BLK_ASG) {
if(b->fromstack == 1) { /* push stack */
if(stackptr == STACKSIZE) eprint("internal function stack overflow");
stack[stackptr++] = fm[0];
stack[stackptr++] = fm[1];
if(icfdebug) puts("pushed");
return 0;
} /* push arg */

b = bnum_to_block(b->gotonum);
fm[2] = b->vartype;
fm[3] = 0;
evalbin(fm+2, TOK_ASG, fm, fm+4);
push_val(b, fm[5]);
if(icfdebug) {
printf("\n%d assigned ",b->bnum);
printval(stdout, fm[5], b->vartype);
nl();
}
return 0;
}

/* not assigned or pushed, we neeed to evaluate */
return cv_bool(fm[1], (short)fm[0]);
} /* eval_form */

static var_in_looparray(bnum)
short bnum;
{
int i;
if(lalist) {
for(i=0; i<lacount; ++i)
if(lalist[i] == bnum) return 1;
}
return 0;
} /* var_in_looparray */

push_val(b,l)
struct block *b;
long l;
{
if(is_array(b)) {
/* is this array part of a running array loop? */
if(var_in_looparray(b->bnum)) {
b->formula[laindex] = l;
} else {
if(b->fmsize == ARRAYLEN)
eprint("1list %s too long, limit %d",
blockname(b), ARRAYLEN);
b->formula[b->fmsize++] = l;
}
} else {
b->answer = l;
}
b->notnull = 1;
} /* push_val */

/* what is the value of a variable?
 * the answer to a question or the result of a calculation. */
var_value(wc, val_ptr, type_ptr)
str_code wc ;/* word code */
long *val_ptr;
short *type_ptr;
{
struct block *b = bname_to_block(wc,0, cbn);
if(!b->notnull)
eprint("variable %s has not ben assigned a value",
blockname(b));
*type_ptr = b->vartype;
if(is_array(b)) {
if(var_in_looparray(b->bnum))
*val_ptr = b->formula[laindex];
else
*val_ptr = b->formula[b->fmsize-1];
} else
*val_ptr = b->answer;
} /* var_value */

printval(outf, l, type)
FILE *outf; /* output file */
long l;
short type;
{
if(type >= 100) {
l = cv_dom_mm(type, (short)l, 0, 0);
type = 4;
}

switch(type) {
case TYPE_MONEY:
fprintf(outf, "$%ld.%02d", l/100, (int)(l%100));
break;
case TYPE_INT:
fprintf(outf, "%ld", l);
break;
case TYPE_STRING:
fprintf(outf, "%s", code_to_string((str_code)l));
break;
case TYPE_BOOL:
fprintf(outf, "%c", l ? 'Y' : 'N');
break;
case TYPE_DATE:
fprintf(outf, "%s", datestring(l));
break;
default:
eprint("@default3[%d] in eval.c", type);
} /* switch */
} /* printval */

/* perform aggrigate operation on array */
static aggrigate(arname, op, lp)
str_code arname;
long *lp;
{
struct block *b = bname_to_block(arname,0, cbn);
char *varname = blockname(b);
short i, type1, type2;
long rc;
long l1, l2;

if(op == TOK_LENGTH) {
lp[0] = 2;
lp[1] = b->fmsize;
return;
}

/* arrange types */
type2 = b->vartype;
if(type2 == 5 && op > TOK_LENGTH)
type2 = 2;
lp[0] = type2;
type1 = 0;
if(type2 >= 100)
type1 = type2, type2 = 4;

/* can't add up strings or domains */
if((type2 == TYPE_DATE || type2 == TYPE_STRING) &&
op > TOK_LENGTH)
eprint("cannot add up the members of %s", varname);

if(!b->notnull) {
if(op != TOK_SUM)
eprint("applying aggrigate %s on empty array %s",
ZC_tok_strings[op], varname);
lp[1] = 0;
return;
}

l2 = b->formula[0]; /* first one */
if(type1) l2 = cv_dom_mm(type1, (short)l2, (str_code)0,0);

for(i=1; i<b->fmsize; ++i) {
l1 = b->formula[i];
if(type1) l1 = cv_dom_mm(type1, (short)l1, (str_code)0,0);
switch(op) {
case TOK_SUM:
case TOK_AVG:
l2 += l1;
break;
case TOK_MIN:
case TOK_MAX:
if(type2 != 4)
rc = l1 - l2;
else
rc = ci_strncmp(code_to_string((str_code)l1), code_to_string((str_code)l2), 0);
if((rc > 0) ^ (op == TOK_MIN))
l2 = l1;
break;
default:
eprint("@default5[%d] in eval.c", op);
} /* switch */
} /* end loop over array */

if(op == TOK_AVG) l2 /= b->fmsize;
if(type1) l2 = cv_dom_mm(type1, 0, (str_code)l2,0);
lp[1] = l2;
return;
} /* aggrigate */

/* perform concatenation */
static concat(sc1, sc2, lp)
str_code sc1, sc2;
long *lp;
{
char *s1 = code_to_string(sc1);
char *s2 = code_to_string(sc2);
char spbuf[300];

if(strlen(s1)+strlen(s2) > 300-2)
eprint("concatenated string is too long");
strcpy(spbuf, s1);
strcat(spbuf, s2);
lp[0] = 4;
lp[1] = string_to_code(spbuf);
} /* concat */

/* get a % character, skipping %% */
static char *getpercchar(obuf, ustring, mval, dval)
char *obuf, *ustring;
long mval ; /* money value */
date dval ; /* date value */
{
char *s, *t, c;
long arg;
int m, d, y;
long md; /* money dollars */
int mc; /* money cents */
static char *mnames[] = {0,
"January","February","March","April","May","June",
"July","August","September","October","November","December"};
static char *mabs[] = {0,
"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"};

md = mval / 100;
mc = mval % 100;
if(dval) {
s = datestring(dval);
s[2] = s[5] = 0;
m = atoi(s);
d = atoi(s+3);
y = atoi(s+6);
}

t = 0;
c = 0;

for(s=ustring; *s; ++s) {
if(*s != '%') continue;
rescan:
c = *++s;
if(!c) break;
if(strchr("mdysac", c))
break;
if(strchr("-.lL0123456789", c))
goto rescan;
}

arg = 0;
if(c) {
t = strchr(s+1, '%');
if(t) *t = 0;

*s = 'd';
switch(c) {
case 'c': arg = mc; break;
case 'd': arg = dval ? (long)d : md; break;
case 'm': arg = m; break;
case 'y': arg = y; break;
case 's': arg = (long) mnames[m]; *s = 's'; break;
case 'a': arg = (long) mabs[m]; *s = 's'; break;
} /* end selecting argument to springs */
}

sprintf(obuf + strlen(obuf), ustring, arg);
/* put ustring back the way you found it */
if(t) *t = '%';
if(c) *s = c;

return t;
} /* getpercchar */

/* perform formatted type conversions */
static usingconv(arg, type, ucode, lp)
long arg;
short type;
str_code ucode;
long *lp;
{
long mval, dval;
char *ustring = code_to_string(ucode); /* the using string */
char spbuf[200];
char *s, a1;

if(strlen(ustring) > 200 - 20)
eprint("string in using clause is too long");
spbuf[0] = 0;
s = ustring;
lp[0] = 4;

if(type >= 100) {
arg = cv_dom_mm(type, (short)arg, (str_code)0, 0);
type = 4;
}

switch(type) {
case TYPE_INT:
sprintf(spbuf, ustring, arg);
break;

case TYPE_STRING:
sprintf(spbuf, ustring, code_to_string((str_code)arg));
break;

case TYPE_BOOL:
a1 = arg ? 'Y' : 'N';
sprintf(spbuf, ustring, a1);
break;

case TYPE_MONEY:
mval = arg;
dval = 0;
case TYPE_DATE:
if(type == TYPE_DATE) dval = arg, mval = 0;
while(s = getpercchar(spbuf, s, mval, dval))   ;
break;

default:
eprint("default7[%d] in eval.c", type);
} /* switch on type */

lp[1] = string_to_code(spbuf);
} /* usingconv */

action()
{
struct block *b; /* current block */
struct block *b1;
short retval; /* was formula true or not? */
FILE *inf, *outf;
int c, i;

setjmp(actjmp);
cbn = 1;
stackptr = 0;
while(cbn < blist.n_nodes) {
b = bnum_to_block(cbn);
if(lalist && cbn == laend) lalist = 0;
if(b->vartype == TYPE_FN) {
/* skip function */
while(b->blktype != BLK_ENDFN) ++b;
cbn = b->bnum+1;
continue;
}

if(icfdebug) printf("exec %d[%d]\n", cbn, b->blktype);
switch(b->blktype) {
case BLK_QUES:
ask_question(b);
case BLK_DECL:
break;

case BLK_ASG:
eval_form(b);
break;

case BLK_IF:
case BLK_WHILE:
retval = 0;
if(b->formula) retval = eval_form(b);
if(!retval) {
cbn = b->gotonum;
if(icfdebug)
printf("block %d[%d]%c jumps to block %d\n",
b->bnum, b->blktype, b->formula ? '*' : ' ', cbn);
continue;
}
break;

case BLK_ARCLEAR:
b = bnum_to_block(b->gotonum);
b->fmsize = 0;
b->notnull = 0;
break;

case BLK_ARPOP:
b = bnum_to_block(b->gotonum);
if(!--b->fmsize) b->notnull = 0;
break;

case BLK_ECHO:
output_text(b->gotonum, b->explain, 1);
break;

case BLK_LA:
if(!lalist) { /* just entering the loop */
lalist = b->formula;
lacount = b->fmsize;
/* check sizes of arrays */
for(i=0; i<b->fmsize; ++i) {
b1 = bnum_to_block((short)lalist[i]);
if(!i) lamax = b1->fmsize;
else if(b1->fmsize != lamax)
eprint("arrays in loop arrray list have different lengths");
} /* loop over arrays */
laindex = -1;
laend = b->gotonum;
}
if(++laindex == lamax) {
cbn = b->gotonum;
continue;
}
break;

case BLK_ENDARG:
if(stack[stackptr-2] != 99)
eprint("too many parameters passed or values returned");
if(b->fromstack) stackptr -= 2;
break;

case BLK_RET:
case BLK_ENDFN:
for(i=stackptr-2; stack[i] != 99; i-=2)  ;
cbn = stack[i+1];
continue;

case BLK_PARMSET:
if(stackptr == STACKSIZE) eprint("internal function stack overflow");
stack[stackptr++] = 99;
stack[stackptr++] = 0;
break;

case BLK_CALL:
for(i=stackptr-2; stack[i] != 99; i-=2)  ;
stack[i+1] = cbn+1;
/* perhaps this is a reserved C function. */
if(exec_c_function((str_code)(b->answer))) break;
b = bname_to_block((str_code)(b->answer),2, cbn);
cbn = b->bnum+1;
continue;

default:
eprint("@default4[%d] in eval.c", b->blktype);
} /* end switch */

++cbn;
} /* end while */
fclose(formfile);

/* cat stream files together to produce the output file */
outf = efopen("icf.out","w");
for(i=1; i<100; ++i) {
if(!streamuse[i]) continue;
inf = efopen(tfname(i), "r");
while((c = getc(inf)) != EOF)
putc(c, outf);
fclose(inf);
}
fclose(outf);

tempclean();

if(opt_f) { /* formatting */
int w = 0, l = 0;
if(opt_w) w = atoi(opt_w);
if(opt_l) l = atoi(opt_l);
mini_mmx("icf.out","icf.fmt",l, w);
} /* formatting */
} /* action */

/* run a C function, rather than an ICF function. */
exec_c_function(fncode)
str_code fncode;
{
char *fname = code_to_string(fncode);
int i, j;
long rv;
long p[6]; /* parameters, at most 6 */
long (*f)();
extern struct {
char *name;
long(*f)();
} c_fns[];

/* find the function */
for(i=0; c_fns[i].name; ++i)
if(!strcmp(fname, c_fns[i].name)) break;
if(!c_fns[i].name) return 0; /* not a reserved C function */
f = c_fns[i].f;

for(i=stackptr-2; stack[i] != 99; i-=2)  ;
if(stackptr-i > (6+1)*2)
eprint("cannot pass more than 6 parameters to a C function");
memset(p, 0, sizeof(p));
j = 0;
while(stack[stackptr-2] != 99) {
p[j] = stack[stackptr-1];
stackptr -= 2;
if(stack[stackptr] == TYPE_STRING)
p[j] = (long) code_to_string((str_code)p[j]);
++j;
}

/* call the C function */
rv = (*f)(p[0], p[1], p[2], p[3], p[4], p[5]);

/* push the (long) return value */
if(stackptr == STACKSIZE) eprint("internal function stack overflow");
stack[stackptr++] = TYPE_INT;
stack[stackptr++] = rv;
return 1;
} /* exec_c_function */

